@@ -1,11 +1,9 @@
 from __future__ import print_function
 
-import platform
 import re
 import sys
 import warnings
 from collections import deque
-from functools import wraps
 from inspect import isclass
 from inspect import isfunction
 from inspect import isgenerator
@@ -13,6 +11,7 @@ from inspect import isgeneratorfunction
 from inspect import ismethod
 from inspect import ismethoddescriptor
 from inspect import isroutine
+from inspect import ismodule
 from logging import getLogger
 
 try:
@@ -25,35 +24,27 @@ try:
 except ImportError:
     ClassType = type
 
+from .utils import basestring
+from .utils import force_bind
+from .utils import make_method_matcher
+from .utils import mimic
+from .utils import PY2
+from .utils import PY3
+from .utils import logf
+from .utils import Sentinel
+
+
 __all__ = 'weave', 'Aspect', 'Proceed', 'Return', 'ALL_METHODS', 'NORMAL_METHODS'
 
 logger = getLogger(__name__)
-
-PY3 = sys.version_info[0] == 3
-PY2 = sys.version_info[0] == 2
-PYPY = platform.python_implementation() == 'PyPy'
-
-if PY3:
-    unicode = str  # pylint: disable=W0622
+logdebug = logf(logger.debug)
+logexception = logf(logger.exception)
 
 
-class _Sentinel(object):
-    def __init__(self, name, doc=''):
-        self.name = name
-        self.__doc__ = doc
-
-    def __repr__(self):
-        if not self.__doc__:
-            return "%s" % self.name
-        else:
-            return "%s: %s" % (self.name, self.__doc__)
-    __str__ = __repr__
-
-UNSPECIFIED = _Sentinel('UNSPECIFIED')
+UNSPECIFIED = Sentinel('UNSPECIFIED')
 ABSOLUTELLY_ALL_METHODS = re.compile('.*')
 ALL_METHODS = re.compile('(?!__getattribute__$)')
 NORMAL_METHODS = re.compile('(?!__.*__$)')
-REGEX_TYPE = type(NORMAL_METHODS)
 VALID_IDENTIFIER = re.compile(r'^[^\W\d]\w*$', re.UNICODE if PY3 else 0)
 
 
@@ -100,22 +91,6 @@ class Return(object):
         self.value = value
 
 
-def mimic(wrapper, func):
-    try:
-        wrapper.__name__ = func.__name__
-    except (TypeError, AttributeError):
-        pass
-    try:
-        wrapper.__module__ = func.__module__
-    except (TypeError, AttributeError):
-        pass
-    try:
-        wrapper.__doc__ = func.__doc__
-    except (TypeError, AttributeError):
-        pass
-    return wrapper
-
-
 class Aspect(object):
     """
     Container for the advice yielding generator. Can be used as a decorator on other function to change behavior
@@ -141,7 +116,7 @@ class Aspect(object):
                     try:
                         advice = next(advisor)
                         while True:
-                            logger.debug('Got advice %r from %s', advice, self.advising_function)
+                            logdebug('Got advice %r from %s', advice, self.advising_function)
                             if advice is Proceed or advice is None or isinstance(advice, Proceed):
                                 if isinstance(advice, Proceed):
                                     args = advice.args
@@ -151,22 +126,22 @@ class Aspect(object):
                                     try:
                                         generated = next(gen)
                                     except StopIteration as exc:
-                                        logger.exception("The cutpoint has been exhausted (early).")
+                                        logexception("The cutpoint has been exhausted (early).")
                                         result = exc.args and exc.args[0]
                                     else:
                                         while True:
                                             try:
                                                 sent = yield generated
                                             except GeneratorExit as exc:
-                                                logger.exception("Got GeneratorExit while consuming the cutpoint")
+                                                logexception("Got GeneratorExit while consuming the cutpoint")
                                                 gen.close()
                                                 raise exc
                                             except BaseException as exc:
-                                                logger.exception("Got exception %r. Throwing it the cutpoint", exc)
+                                                logexception("Got exception %r. Throwing it the cutpoint", exc)
                                                 try:
                                                     generated = gen.throw(*sys.exc_info())
                                                 except StopIteration as exc:
-                                                    logger.exception("The cutpoint has been exhausted.")
+                                                    logexception("The cutpoint has been exhausted.")
                                                     result = exc.args and exc.args[0]
                                                     break
                                             else:
@@ -176,7 +151,7 @@ class Aspect(object):
                                                     else:
                                                         generated = gen.send(sent)
                                                 except StopIteration as exc:
-                                                    logger.exception("The cutpoint has been exhausted.")
+                                                    logexception("The cutpoint has been exhausted.")
                                                     result = exc.args and exc.args[0]
                                                     break
                                 except BaseException as exc:
@@ -205,7 +180,7 @@ class Aspect(object):
                 try:
                     advice = next(advisor)
                     while True:
-                        logger.debug('Got advice %r from %s', advice, self.advising_function)
+                        logdebug('Got advice %r from %s', advice, self.advising_function)
                         if advice is Proceed or advice is None or isinstance(advice, Proceed):
                             if isinstance(advice, Proceed):
                                 args = advice.args
@@ -229,42 +204,6 @@ class Aspect(object):
                     advisor.close()
             return mimic(advising_function_wrapper, cutpoint_function)
 
-#class Boundable(object):
-#    def __init__(self, binding=None):
-#        self.binding = binding
-#
-#    def __get__(self, instance, owner):
-#        return type(self)(instance)
-
-#class DescriptorAspect(Aspect):
-#    """
-#    Aspect that call the generator with ``instance, args, kwargs, cutpoint`` instead of the original arguments.
-#    """
-#    def __init__(self, advising_function):
-#        if not isgeneratorfunction(advising_function):
-#            raise ExpectedGeneratorFunction("advising_function %s must be a generator function." % advising_function)
-#        self.advising_function = advising_function
-#
-#    def __call__(self, cutpoint_function):
-#        wrapper = super(DescriptorAspect, self).__call__(lambda _i, args, kwargs, _f: cutpoint_function(*args, **kwargs))
-#        cutpoint_wrapping = wraps(cutpoint_function)
-#
-#        class descriptor_wrapper(object):
-#            __doc__ = cutpoint_function.__doc__
-#            __name__ = cutpoint_function.__name__
-#            __module__ = cutpoint_function.__module__
-#
-#            def __get__(self, instance, owner):
-#                print('__get__(', instance, owner, ')')
-#                return cutpoint_wrapping(lambda *args, **kwargs:
-#                    wrapper(instance, args, kwargs, cutpoint_function)
-#                )
-#
-#            def __call__(self, *args, **kwargs):
-#                print('__call__(', args, kwargs, ')')
-#                return wrapper(None, args, kwargs, cutpoint_function)
-#
-#        return descriptor_wrapper()
 
 class Fabric(object):
     pass
@@ -284,8 +223,8 @@ class Rollback(object):
         else:
             self._rollbacks = [rollback]
 
-    def merge(self, other):
-        self._rollbacks.append(other)
+    def merge(self, *others):
+        self._rollbacks.extend(others)
 
     def __enter__(self):
         return self
@@ -297,8 +236,8 @@ class Rollback(object):
     rollback = __call__ = __exit__
 
 
-def checked_apply(aspects, function):
-    logger.debug('Applying aspects %s to function %s.', aspects, function)
+def _checked_apply(aspects, function, module=None):
+    logdebug(' > applying aspects %s to function %s.', aspects, function)
     if callable(aspects):
         wrapper = aspects(function)
         assert callable(wrapper), 'Aspect %s did not return a callable (it return %s).' % (aspects, wrapper)
@@ -307,10 +246,10 @@ def checked_apply(aspects, function):
         for aspect in aspects:
             wrapper = aspect(wrapper)
             assert callable(wrapper), 'Aspect %s did not return a callable (it return %s).' % (aspect, wrapper)
-    return wrapper
+    return mimic(wrapper, function, module=module)
 
 
-def check_name(name):
+def _check_name(name):
     if not VALID_IDENTIFIER.match(name):
         raise SyntaxError(
             "Could not match %r to %r. It should be a string of "
@@ -337,8 +276,8 @@ def weave(target, aspects, **options):
         If ``True``, aliases of target are replaced.
 
     :param bool lazy:
-        If ``True`` only patch target's ``__init__``, the rest of the methods are patched after ``__init__`` is
-        called. *Only available for classes*
+        If ``True`` only target's ``__init__`` method is patched, the rest of the methods are patched after ``__init__``
+        is called. *Only available for classes*.
 
     :param methods: Methods from target to patch. *Only available for classes*
     :type methods: list or regex or string
@@ -349,6 +288,13 @@ def weave(target, aspects, **options):
     :raises TypeError:
         If target is a unacceptable object, or the specified options are not available for that type of object.
 
+
+    .. versionchanged:: 0.4.0
+
+        Replaced `only_methods`, `skip_methods`, `skip_magicmethods` options with `methods`.
+        Renamed `on_init` option to `lazy`.
+        Added `aliases` option.
+        Replaced `skip_subclasses` option with `subclasses`.
     """
     if not callable(aspects):
         if not hasattr(aspects, '__iter__'):
@@ -361,11 +307,14 @@ def weave(target, aspects, **options):
         return Rollback([
             weave(item, aspects, **options) for item in target
         ])
-    elif isinstance(target, (unicode, str)):
-        assert '.' in target, "Need at least a module in the target specification !"
+    elif isinstance(target, basestring):
         parts = target.split('.')
         for part in parts:
-            check_name(part)
+            _check_name(part)
+
+        if len(parts) == 1:
+            __import__(part)
+            return weave_module(sys.modules[part], aspects, **options)
 
         for pos in reversed(range(1, len(parts))):
             owner, name = '.'.join(parts[:pos]), '.'.join(parts[pos:])
@@ -385,29 +334,31 @@ def weave(target, aspects, **options):
             while path:
                 owner = getattr(owner, path.popleft())
 
-        logger.debug("Patching %s from %s ...", name, owner)
+        logdebug(" ~ patching %s from %s ...", name, owner)
         obj = getattr(owner, name)
+
         if isinstance(obj, (type, ClassType)):
-            logger.debug(" .. as a class %r.", obj)
+            logdebug("   .. as a class %r.", obj)
             return weave_class(
                 obj, aspects,
                 owner=owner, name=name, **options
             )
         elif callable(obj):  # or isinstance(obj, FunctionType) ??
-            logger.debug(" .. as a callable %r.", obj)
+            logdebug("   .. as a callable %r.", obj)
             return weave_module_function(owner, obj, aspects, force_name=name, **options)
         else:
-            raise TypeError("Can't weave object %s of type %s" % (obj, type(obj)))
+            return weave(obj, aspects, **options)
     name = getattr(target, '__name__', None)
+    logdebug("weave (target=%s, aspects=%s, name=%s, **options=%s)", target, aspects, name, options)
     if name and getattr(__builtin__, name, None) is target:
         return weave_module_function(__builtin__, target, aspects, **options)
     elif PY3 and ismethod(target):
         inst = target.__self__
         name = target.__name__
-        logger.debug("Weaving %r (%s) as instance method.", target, name)
+        logdebug(" ~ patching %r (%s) as instance method.", target, name)
         assert not options, "keyword arguments are not supported when weaving instance methods."
         func = getattr(inst, name)
-        setattr(inst, name, checked_apply(aspects, func).__get__(inst, type(inst)))
+        setattr(inst, name, _checked_apply(aspects, func).__get__(inst, type(inst)))
         return Rollback(lambda: delattr(inst, name))
     elif PY3 and isfunction(target):
         owner = __import__(target.__module__)
@@ -415,19 +366,19 @@ def weave(target, aspects, **options):
         while path:
             owner = getattr(owner, path.popleft())
         name = target.__name__
-        logger.debug("Weaving %r (%s) as a property.", target, name)
+        logdebug(" ~ patching %r (%s) as a property.", target, name)
         func = owner.__dict__[name]
-        return patch_module(owner, name, checked_apply(aspects, func), func, **options)
+        return patch_module(owner, name, _checked_apply(aspects, func), func, **options)
     elif PY2 and isfunction(target):
         return weave_module_function(__import__(target.__module__), target, aspects, **options)
     elif PY2 and ismethod(target):
         if target.im_self:
             inst = target.im_self
             name = target.__name__
-            logger.debug("Weaving %r (%s) as instance method.", target, name)
+            logdebug(" ~ patching %r (%s) as instance method.", target, name)
             assert not options, "keyword arguments are not supported when weaving instance methods."
             func = getattr(inst, name)
-            setattr(inst, name, checked_apply(aspects, func).__get__(inst, type(inst)))
+            setattr(inst, name, _checked_apply(aspects, func).__get__(inst, type(inst)))
             return Rollback(lambda: delattr(inst, name))
         else:
             klass = target.im_class
@@ -435,74 +386,133 @@ def weave(target, aspects, **options):
             return weave(klass, aspects, methods='%s$' % name, **options)
     elif isclass(target):
         return weave_class(target, aspects, **options)
+    elif ismodule(target):
+        return weave_module(target, aspects, **options)
+    elif type(target).__module__ not in ('builtins', '__builtin__'):
+        return weave_instance(target, aspects, **options)
     else:
         raise UnsupportedType("Can't weave object %s of type %s" % (target, type(target)))
 
 
-def rewrap_method(func, klass, aspect):
+def _rewrap_method(func, klass, aspect):
     if isinstance(func, staticmethod):
         if hasattr(func, '__func__'):
-            return staticmethod(checked_apply(aspect, func.__func__))
+            return staticmethod(_checked_apply(aspect, func.__func__))
         else:
-            return staticmethod(checked_apply(aspect, func.__get__(None, klass)))
+            return staticmethod(_checked_apply(aspect, func.__get__(None, klass)))
     elif isinstance(func, classmethod):
         if hasattr(func, '__func__'):
-            return classmethod(checked_apply(aspect, func.__func__))
+            return classmethod(_checked_apply(aspect, func.__func__))
         else:
-            return classmethod(checked_apply(aspect, func.__get__(None, klass).im_func))
+            return classmethod(_checked_apply(aspect, func.__get__(None, klass).im_func))
     else:
-        return checked_apply(aspect, func)
+        return _checked_apply(aspect, func)
+
+
+def weave_instance(instance, aspect, methods=NORMAL_METHODS, lazy=False, **options):
+    """
+    Low-level weaver for instances.
+
+    .. warning:: You should not use this directly.
+
+    :returns: An :obj:`aspectlib.Entanglement` object.
+    """
+    entanglement = Rollback()
+    method_matches = make_method_matcher(methods)
+    logdebug("weave_instance (module=%r, aspect=%s, methods=%s, lazy=%s, **options=%s)",
+             instance, aspect, methods, lazy, options)
+    fixup = lambda func: func.__get__(instance, type(instance))
+    fixed_aspect = aspect + [fixup] if isinstance(aspect, (list, tuple)) else [aspect, fixup]
+
+    for attr in dir(instance):
+        func = getattr(instance, attr)
+        if method_matches(attr):
+            if ismethod(func):
+                logger.info("%s %s %s", attr, func, type(func))
+                if hasattr(func, '__func__'):
+                    realfunc = func.__func__
+                else:
+                    realfunc = func.im_func
+                entanglement.merge(
+                    patch_module(instance, attr, _checked_apply(fixed_aspect, realfunc, module=None), **options)
+                )
+    return entanglement
+
+def weave_module(module, aspect, methods=NORMAL_METHODS, lazy=False, **options):
+    """
+    Low-level weaver for "whole module weaving".
+
+    .. warning:: You should not use this directly.
+
+    :returns: An :obj:`aspectlib.Entanglement` object.
+    """
+    entanglement = Rollback()
+    method_matches = make_method_matcher(methods)
+    logdebug("weave_module (module=%r, aspect=%s, methods=%s, lazy=%s, **options=%s)",
+             module, aspect, methods, lazy, options)
+
+    for attr in dir(module):
+        func = getattr(module, attr)
+        if method_matches(attr):
+            if isroutine(func):
+                entanglement.merge(weave_module_function(module, func, aspect, force_name=attr, **options))
+            elif isclass(func):
+                entanglement.merge(
+                    weave_class(func, aspect, owner=module, name=attr, methods=methods, lazy=lazy, **options),
+                    #  it's not consistent with the other ways of weaving a class (it's never weaved as a routine).
+                    #  therefore it's disabled until it's considered useful.
+                    #weave_module_function(module, getattr(module, attr), aspect, force_name=attr, **options),
+                )
+    return entanglement
 
 
 def weave_class(klass, aspect, methods=NORMAL_METHODS, subclasses=True, lazy=False,
                 owner=None, name=None, aliases=True):
+    """
+    Low-level weaver for classes.
+
+    .. warning:: You should not use this directly.
+    """
 
     assert isclass(klass), "Can't weave %r. Must be a class." % klass
     entanglement = Rollback()
-    if isinstance(methods, (str, unicode)):
-        method_matches = re.compile(methods).match
-    elif isinstance(methods, (list, tuple)):
-        method_matches = methods.__contains__
-    elif isinstance(methods, REGEX_TYPE):
-        method_matches = methods.match
-    else:
-        raise TypeError("Unacceptable methods spec %r." % methods)
+    method_matches = make_method_matcher(methods)
 
     if subclasses and hasattr(klass, '__subclasses__'):
         for sub_class in klass.__subclasses__():
             if not issubclass(sub_class, Fabric):
                 entanglement.merge(weave_class(sub_class, aspect, methods=methods, subclasses=subclasses, lazy=lazy))
+    logdebug("weave_class (klass=%r, methods=%s, subclasses=%s, lazy=%s, owner=%s, name=%s, aliases=%s)",
+          klass, methods, subclasses, lazy, owner, name, aliases)
     if lazy:
-        logger.debug("Weaving %r as class (on demand at __init__ time).", klass)
-
         def __init__(self, *args, **kwargs):
             super(SubClass, self).__init__(*args, **kwargs)
             for attr in dir(self):
                 func = getattr(self, attr, None)
                 if method_matches(attr) and attr not in wrappers and isroutine(func):
-                    setattr(self, attr, checked_apply(aspect, force_bind(func)).__get__(self, SubClass))
+                    setattr(self, attr, _checked_apply(aspect, force_bind(func)).__get__(self, SubClass))
 
         wrappers = {
-            '__init__': checked_apply(aspect, __init__) if method_matches('__init__') else __init__
+            '__init__': _checked_apply(aspect, __init__) if method_matches('__init__') else __init__
         }
         for attr, func in klass.__dict__.items():
             if method_matches(attr):
                 if ismethoddescriptor(func):
-                    wrappers[attr] = rewrap_method(func, klass, aspect)
+                    wrappers[attr] = _rewrap_method(func, klass, aspect)
 
-        logger.debug(" * Creating subclass with attributes %r", wrappers)
+        logdebug(" * creating subclass with attributes %r", wrappers)
         name = name or klass.__name__
         SubClass = type(name, (klass, Fabric), wrappers)
         SubClass.__module__ = klass.__module__
         module = owner or __import__(klass.__module__)
         entanglement.merge(patch_module(module, name, SubClass, original=klass, aliases=aliases))
     else:
-        logger.debug("Weaving %r as class.", klass)
         original = {}
         for attr, func in klass.__dict__.items():
             if method_matches(attr):
                 if isroutine(func):
-                    setattr(klass, attr, rewrap_method(func, klass, aspect))
+                    logdebug(" ~ patching attributes %r (original: %r).", attr, func)
+                    setattr(klass, attr, _rewrap_method(func, klass, aspect))
                 else:
                     continue
                 original[attr] = func
@@ -514,11 +524,22 @@ def weave_class(klass, aspect, methods=NORMAL_METHODS, subclasses=True, lazy=Fal
     return entanglement
 
 
-def patch_module(module, name, replacement, original=UNSPECIFIED, aliases=True):
+def patch_module(module, name, replacement, original=UNSPECIFIED, aliases=True, location=None):
+    """
+    Low-level attribute patcher.
+
+    :param module module: Object to patch.
+    :param str name: Attribute to patch
+    :param replacement: The replacement value.
+    :param original: The original value (in case the object beeing patched uses descriptors or is plain weird).
+    :param bool aliases: If ``True`` patch all the attributes that have the same original value.
+
+    :returns: An :obj:`aspectlib.Entanglement` object.
+    """
     rollback = Rollback()
     seen = False
     original = getattr(module, name) if original is UNSPECIFIED else original
-    location = module.__name__
+    location = module.__name__ if hasattr(module, '__name__') else type(module).__module__
     try:
         replacement.__module__ = location
     except (TypeError, AttributeError):
@@ -528,14 +549,14 @@ def patch_module(module, name, replacement, original=UNSPECIFIED, aliases=True):
             obj = getattr(module, alias)
             if obj is original:
                 if aliases or alias == name:
-                    logger.debug(" * Saving %s on %s.%s ...", replacement, location, alias)
+                    logdebug(" ~ saving %s on %s.%s ...", replacement, location, alias)
                     setattr(module, alias, replacement)
                     rollback.merge(lambda alias=alias: setattr(module, alias, original))
                 if alias == name:
                     seen = True
             elif alias == name:
                 if ismethod(obj):
-                    logger.debug(" * Saving %s on %s.%s ...", replacement, location, alias)
+                    logdebug(" ~ saving %s on %s.%s ...", replacement, location, alias)
                     setattr(module, alias, replacement)
                     rollback.merge(lambda alias=alias: setattr(module, alias, original))
                 else:
@@ -545,21 +566,21 @@ def patch_module(module, name, replacement, original=UNSPECIFIED, aliases=True):
         warnings.warn('Setting %s.%s to %s. There was no previous definition, probably patching the wrong module.' % (
             location, name, replacement
         ))
-        logger.debug(" * Saving %s on %s.%s ...", replacement, location, name)
+        logdebug(" ~ saving %s on %s.%s ...", replacement, location, name)
         setattr(module, name, replacement)
         rollback.merge(lambda: setattr(module, name, original))
     return rollback
 
 
-def force_bind(func):
-    def bound(self, *args, **kwargs):  # pylint: disable=W0613
-        return func(*args, **kwargs)
-    bound.__name__ = func.__name__
-    bound.__doc__ = func.__doc__
-    return bound
+def weave_module_function(module, target, aspect, force_name=None, **options):
+    """
+    Low-level weaver for one function from a specified module.
 
+    .. warning:: You should not use this directly.
 
-def weave_module_function(mod, target, aspect, force_name=None, **options):
-    logger.debug("Weaving %r as plain function.", target)
+    :returns: An :obj:`aspectlib.Entanglement` object.
+    """
+    logdebug("weave_module_function (module=%s, target=%s, aspect=%s, force_name=%s, **options=%s",
+          module, target, aspect, force_name, options)
     name = force_name or target.__name__
-    return patch_module(mod, name, checked_apply(aspect, target), original=target, **options)
+    return patch_module(module, name, _checked_apply(aspect, target, module=module), original=target, **options)
