@@ -378,7 +378,7 @@ def test_weave_str_class_meth_target():
 
 def test_weave_wrong_module():
     calls = []
-    with aspectlib.weave('warnings.warn', record(history=calls)):
+    with aspectlib.weave('warnings.warn', record(calls=calls)):
         aspectlib.weave(AliasedGlobal, mock('stuff'), lazy=True)
     assert calls == [
         (None,
@@ -481,7 +481,7 @@ def test_weave_instance_meth():
     assert inst.foo == 'stuff'
 
 
-@pytest.mark.xfail
+@pytest.mark.skipif('aspectlib.PY2')
 def test_weave_legacy_instance():
     @aspectlib.Aspect
     def aspect(self):
@@ -497,7 +497,6 @@ def test_weave_legacy_instance():
     assert inst.foo == 'stuff'
 
 
-@pytest.mark.xfail
 def test_weave_instance():
     @aspectlib.Aspect
     def aspect(self):
@@ -1081,7 +1080,7 @@ def test_unspecified_str():
 
 
 def test_sentinel():
-    assert repr(aspectlib._Sentinel('STUFF', "Means it's some stuff")) == "STUFF: Means it's some stuff"
+    assert repr(aspectlib.Sentinel('STUFF', "Means it's some stuff")) == "STUFF: Means it's some stuff"
 
 
 def _internal():
@@ -1106,7 +1105,7 @@ else:
 
 
 def test_invalid_string_target():
-    raises(AssertionError, aspectlib.weave, 'inva lid', mock(None))
+    raises(SyntaxError, aspectlib.weave, 'inva lid', mock(None))
     raises(SyntaxError, aspectlib.weave, 'os.inva lid', mock(None))
     raises(SyntaxError, aspectlib.weave, 'os.2invalid', mock(None))
     raises(SyntaxError, aspectlib.weave, 'os.some,junk', mock(None))
@@ -1118,11 +1117,11 @@ def test_invalid_string_target():
 
 
 def test_list_of_aspects():
-    with aspectlib.weave(module_func, [mock('foobar'), record(call=True)]):
+    with aspectlib.weave(module_func, [mock('foobar'), record]):
         assert module_func(1, 2, 3) == 'foobar'
         assert module_func.calls == [(None, (1, 2, 3), {})]
 
-    with aspectlib.weave(module_func, [mock('foobar', call=True), record(call=True)]):
+    with aspectlib.weave(module_func, [mock('foobar', call=True), record]):
         raises(TypeError, module_func, 1, 2, 3)
         assert module_func.calls == [(None, (1, 2, 3), {})]
 
@@ -1418,7 +1417,19 @@ def test_aspect_on_coroutine():
     print (hist)
     assert hist == ['before', 'the-return-value', 'after', 'finally', 'closed']
 
-if __name__ == '__main__':
-    test_aspect_on_generator_func_bad_advice()
-    import os
-    os._exit(1)
+
+
+def test_weave_module(strmod=None):
+    calls = []
+    from test_pkg1.test_pkg2 import test_mod
+    with aspectlib.weave(strmod or test_mod, record(calls=calls, extended=True)):
+        test_mod.target()
+        obj = test_mod.Stuff()
+        obj.meth()
+    assert calls == [
+        (None, 'test_pkg1.test_pkg2.test_mod.target', (), {}),
+        (obj, 'test_pkg1.test_pkg2.test_mod.meth', (), {})
+    ]
+
+def test_weave_module_as_str():
+    test_weave_module("test_pkg1.test_pkg2.test_mod")
